#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
from .utils import Asn1ToolsBaseTest
import asn1tools
import sys
from copy import deepcopy
from asn1tools.codecs import utc_time_to_datetime as ut2dt
from asn1tools.codecs import generalized_time_to_datetime as gt2dt

sys.path.append('tests/files/3gpp')

from rrc_8_6_0 import EXPECTED as RRC_8_6_0


class Asn1ToolsXerTest(Asn1ToolsBaseTest):

    maxDiff = None

    def test_real(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= REAL "
            "END",
            'xer')

        datas = [
            ('A', 1.0, b'<A>1.0E0</A>'),
            ('A', 10.0, b'<A>1.0E1</A>'),
            ('A', -1.0, b'<A>-1.0E0</A>'),
            ('A', -2, b'<A>-2.0E0</A>'),
            ('A', -9.99, b'<A>-9.99E0</A>'),
            ('A', -10.0, b'<A>-1.0E1</A>'),
            ('A', 0.31, b'<A>0.31E0</A>'),
            ('A', 2000.0, b'<A>2.0E3</A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_null(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= NULL "
            "B ::= SEQUENCE OF NULL "
            "C ::= SEQUENCE { a NULL } "
            "END",
            'xer')

        datas = [
            ('A',         None, b'<A />'),
            ('B', [None, None], b'<B><NULL /><NULL /></B>'),
            ('C',  {'a': None}, b'<C><a /></C>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_bit_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= BIT STRING "
            "END",
            'xer')

        datas = [
            ('A',         (b'', 0), b'<A />'),
            ('A', (b'\x00', 2), b'<A>00</A>'),
            ('A',     (b'\x40', 4), b'<A>0100</A>'),
            ('A', (b'\x40\x80', 9), b'<A>010000001</A>'),
            ('A',
             (b'\x11\x22\x33\x44\x55\x66\x76', 55),
             b'<A>0001000100100010001100110100010001010101011001100111011</A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_octet_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= OCTET STRING "
            "END",
            'xer')

        datas = [
            ('A',         b'', b'<A />'),
            ('A', b'\x40\x80', b'<A>4080</A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_object_identifier(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= OBJECT IDENTIFIER "
            "END",
            'xer')

        datas = [
            ('A',    '1.2.3', b'<A>1.2.3</A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode('A', b'<A></A>')

        self.assertEqual(str(cm.exception),
                         "Expected an OBJECT IDENTIFIER, but got ''.")

    def test_external(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= EXTERNAL "
            "END",
            'xer')

        datas = [
            ('A',
             {'encoding': ('octet-aligned', b'\x12')},
             b'<A><encoding><octet-aligned>12</octet-aligned></encoding></A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_enumerated(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= ENUMERATED { r(5), t(10) } "
            "B ::= SEQUENCE OF ENUMERATED { a(0) } "
            "END",
            'xer')

        datas = [
            ('A',    'r', b'<A><r /></A>'),
            ('A',    't', b'<A><t /></A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

        # Encode error.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            foo.encode('A', 'foo')

        self.assertEqual(
            str(cm.exception),
            "Expected enumeration value 'r' or 't', but got 'foo'.")

        # Decode error.
        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode('A', b'<A><bar /></A>')

        self.assertEqual(
            str(cm.exception),
            "Expected enumeration value 'r' or 't', but got 'bar'.")

        # Encode of error.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            foo.encode('B', ['foo'])

        self.assertEqual(
            str(cm.exception),
            "Expected enumeration value 'a', but got 'foo'.")

        # Decode of error.
        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode('B', b'<B><bar /></B>')

        self.assertEqual(
            str(cm.exception),
            "Expected enumeration value 'a', but got 'bar'.")

    def test_sequence(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SEQUENCE { "
            "  a SEQUENCE OF A OPTIONAL "
            "} "
            "B ::= SEQUENCE { "
            "  a INTEGER DEFAULT 4 "
            "} "
            "END",
            'xer')

        datas = [
            ('A',                    {}, b'<A />'),
            ('A',           {'a': [{}]}, b'<A><a><A /></a></A>'),
            ('A',    {'a': [{'a': []}]}, b'<A><a><A><a /></A></a></A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

        # Non-symmetrical encoding and decoding.
        self.assertEqual(foo.encode('B', {}), b'<B />')
        self.assertEqual(foo.decode('B', b'<B />'), {'a': 4})

    def test_sequence_of(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SEQUENCE OF INTEGER "
            "B ::= SEQUENCE OF A "
            "C ::= SEQUENCE OF BOOLEAN "
            "D ::= SEQUENCE OF E "
            "E ::= BOOLEAN "
            "F ::= SEQUENCE OF CHOICE { a BOOLEAN, b INTEGER } "
            "G ::= SEQUENCE OF ENUMERATED { one } "
            "H ::= SEQUENCE OF SEQUENCE { a INTEGER } "
            "I ::= SEQUENCE OF BIT STRING "
            "J ::= SEQUENCE OF OCTET STRING "
            "K ::= SEQUENCE OF OBJECT IDENTIFIER "
            "L ::= SEQUENCE OF SEQUENCE OF NULL "
            "M ::= SEQUENCE OF SET OF NULL "
            "END",
            'xer')

        datas = [
            ('A',               [], b'<A />'),
            ('A',           [1, 4], b'<A><INTEGER>1</INTEGER><INTEGER>4</INTEGER></A>'),
            ('B',            [[5]], b'<B><A><INTEGER>5</INTEGER></A></B>'),
            ('C',    [True, False], b'<C><true /><false /></C>'),
            ('D',           [True], b'<D><true /></D>'),
            ('F',
             [('a', True), ('b', 1)],
             b'<F><a><true /></a><b>1</b></F>'),
            ('G',          ['one'], b'<G><one /></G>'),
            ('H',       [{'a': 1}], b'<H><SEQUENCE><a>1</a></SEQUENCE></H>'),
            ('I',   [(b'\x80', 1)], b'<I><BIT_STRING>1</BIT_STRING></I>'),
            ('J',        [b'\x12'], b'<J><OCTET_STRING>12</OCTET_STRING></J>'),
            ('K',
             ['1.2.30'],
             b'<K><OBJECT_IDENTIFIER>1.2.30</OBJECT_IDENTIFIER></K>'),
            ('L',
             [[None], []],
             b'<L><SEQUENCE_OF><NULL /></SEQUENCE_OF><SEQUENCE_OF /></L>'),
            ('M',
             [[None], []],
             b'<M><SET_OF><NULL /></SET_OF><SET_OF /></M>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_choice(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= CHOICE { "
            "  a BOOLEAN, "
            "  b INTEGER, "
            "  c CHOICE { "
            "    a INTEGER "
            "  } "
            "} "
            "B ::= SEQUENCE OF A "
            "END",
            'xer')

        datas = [
            ('A', ('a', True), b'<A><a><true /></a></A>'),
            ('A', ('b', 1), b'<A><b>1</b></A>'),
            ('A', ('c', ('a', 534)), b'<A><c><a>534</a></c></A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

        # Encode error.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            foo.encode('A', ('d', None))

        self.assertEqual(str(cm.exception),
                         "Expected choice 'a', 'b' or 'c', but got 'd'.")

        # Decode error.
        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode('A', b'<A><d><true /></d></A>')

        self.assertEqual(str(cm.exception),
                         "Expected choice 'a', 'b' or 'c', but got 'd'.")

        # Encode of error.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            foo.encode('B', [('d', None)])

        self.assertEqual(str(cm.exception),
                         "Expected choice 'a', 'b' or 'c', but got 'd'.")

        # Decode of error.
        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode('B', b'<A><d><true /></d></A>')

        self.assertEqual(str(cm.exception),
                         "Expected choice 'a', 'b' or 'c', but got 'd'.")

    def test_utf8_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UTF8String "
            "END",
            'xer')

        datas = [
            ('A',         u'', b'<A />'),
            ('A',      u'bar', b'<A>bar</A>'),
            ('A', u'a\u1010c', b'<A>a&#4112;c</A>'),
            ('A',    u'f → ∝', b'<A>f &#8594; &#8733;</A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_universal_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UniversalString "
            "END",
            'xer')

        datas = [
            ('A',        u'bar', b'<A>bar</A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_utc_time(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UTCTime "
            "END",
            'xer')

        datas = [
            ('A', ut2dt('9205210000Z'), b'<A>9205210000Z</A>'),
            ('A', ut2dt('920622123421Z'), b'<A>920622123421Z</A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_generalized_time(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= GeneralizedTime "
            "END",
            'xer')

        datas = [
            ('A', gt2dt('199205210000Z'), b'<A>199205210000Z</A>'),
            ('A', gt2dt('19920622123421Z'), b'<A>19920622123421Z</A>'),
            ('A', gt2dt('199207221321.3Z'), b'<A>199207221321.3Z</A>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_any(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS ::= "
            "BEGIN "
            "A ::= ANY "
            "END",
            'xer')

        with self.assertRaises(NotImplementedError) as cm:
            foo.encode('A', b'')

        self.assertEqual(str(cm.exception), 'ANY is not yet implemented.')

        with self.assertRaises(NotImplementedError) as cm:
            foo.decode('A', b'<A />')

        self.assertEqual(str(cm.exception), 'ANY is not yet implemented.')

    def test_foo(self):
        foo = asn1tools.compile_files(['tests/files/foo.asn'], 'xer')

        self.assertEqual(len(foo.types), 2)
        self.assertTrue(foo.types['Question'] is not None)
        self.assertTrue(foo.types['Answer'] is not None)
        self.assertEqual(len(foo.modules), 1)
        self.assertTrue(foo.modules['Foo'] is not None)

        # Question.
        decoded_message = {'id': 1, 'question': 'Is 1+1=3?'}
        encoded_message = (
            b'<Question><id>1</id><question>Is 1+1=3?</question></Question>'
        )

        encoded = foo.encode('Question', decoded_message)
        self.assertEqual(encoded, encoded_message)
        decoded = foo.decode('Question', encoded)
        self.assertEqual(decoded, decoded_message)

        # Answer.
        decoded_message = {'id': 1, 'answer': False}
        encoded_message = (
            b'<Answer><id>1</id><answer><false /></answer></Answer>'
        )

        encoded = foo.encode('Answer', decoded_message)
        self.assertEqual(encoded, encoded_message)
        decoded = foo.decode('Answer', encoded)
        self.assertEqual(decoded, decoded_message)

        # Encode a question with missing field 'id'.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            encoded = foo.encode('Question', {'question': 'Is 1+1=3?'})

        self.assertEqual(
            str(cm.exception),
            "Sequence member 'id' not found in {'question': 'Is 1+1=3?'}.")

    def test_decode_length(self):
        foo = asn1tools.compile_files('tests/files/foo.asn', 'xer')

        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode_length(b'')

        self.assertEqual(str(cm.exception),
                         'Decode length is not supported for this codec.')

    def test_rrc_8_6_0(self):
        rrc = asn1tools.compile_dict(deepcopy(RRC_8_6_0), 'xer')

        # Message 1.
        decoded = {
            'message': (
                'c1',
                (
                    'paging',
                    {
                        'systemInfoModification': 'true',
                        'nonCriticalExtension': {
                        }
                    }
                )
            )
        }

        encoded = (
            b'<PCCH-Message><message><c1><paging><systemInfoModification><'
            b'true /></systemInfoModification><nonCriticalExtension /></pa'
            b'ging></c1></message></PCCH-Message>'
        )

        self.assert_encode_decode_string(rrc, 'PCCH-Message', decoded, encoded)

        # Message 2.
        decoded = {
            'message': {
                'dl-Bandwidth': 'n6',
                'phich-Config': {
                    'phich-Duration': 'normal',
                    'phich-Resource': 'half'
                },
                'systemFrameNumber': (b'\x12', 8),
                'spare': (b'\x34\x40', 10)
            }
        }

        encoded = (
            b'<BCCH-BCH-Message><message><dl-Bandwidth><n6 /></dl-Bandwidt'
            b'h><phich-Config><phich-Duration><normal /></phich-Duration><'
            b'phich-Resource><half /></phich-Resource></phich-Config><syst'
            b'emFrameNumber>00010010</systemFrameNumber><spare>0011010001<'
            b'/spare></message></BCCH-BCH-Message>'
        )

        self.assert_encode_decode_string(rrc, 'BCCH-BCH-Message', decoded, encoded)

        # Message 3.
        decoded = {
            'message': (
                'c1',
                (
                    'systemInformation',
                    {
                        'criticalExtensions': (
                            'systemInformation-r8',
                            {
                                'sib-TypeAndInfo': [
                                    (
                                        'sib2',
                                        {
                                            'ac-BarringInfo': {
                                                'ac-BarringForEmergency': True,
                                                'ac-BarringForMO-Data': {
                                                    'ac-BarringFactor': 'p95',
                                                    'ac-BarringTime': 's128',
                                                    'ac-BarringForSpecialAC': (b'\xf0', 5)
                                                }
                                            },
                                            'radioResourceConfigCommon': {
                                                'rach-ConfigCommon': {
                                                    'preambleInfo': {
                                                        'numberOfRA-Preambles': 'n24',
                                                        'preamblesGroupAConfig': {
                                                            'sizeOfRA-PreamblesGroupA': 'n28',
                                                            'messageSizeGroupA': 'b144',
                                                            'messagePowerOffsetGroupB': 'minusinfinity'
                                                        }
                                                    },
                                                    'powerRampingParameters': {
                                                        'powerRampingStep': 'dB0',
                                                        'preambleInitialReceivedTargetPower': 'dBm-102'
                                                    },
                                                    'ra-SupervisionInfo': {
                                                        'preambleTransMax': 'n8',
                                                        'ra-ResponseWindowSize': 'sf6',
                                                        'mac-ContentionResolutionTimer': 'sf48'
                                                    },
                                                    'maxHARQ-Msg3Tx': 8
                                                },
                                                'bcch-Config': {
                                                    'modificationPeriodCoeff': 'n2'
                                                },
                                                'pcch-Config': {
                                                    'defaultPagingCycle': 'rf256',
                                                    'nB': 'twoT'
                                                },
                                                'prach-Config': {
                                                    'rootSequenceIndex': 836,
                                                    'prach-ConfigInfo': {
                                                        'prach-ConfigIndex': 33,
                                                        'highSpeedFlag': False,
                                                        'zeroCorrelationZoneConfig': 10,
                                                        'prach-FreqOffset': 64
                                                    }
                                                },
                                                'pdsch-ConfigCommon': {
                                                    'referenceSignalPower': -60,
                                                    'p-b': 2
                                                },
                                                'pusch-ConfigCommon': {
                                                    'pusch-ConfigBasic': {
                                                        'n-SB': 1,
                                                        'hoppingMode': 'interSubFrame',
                                                        'pusch-HoppingOffset': 10,
                                                        'enable64QAM': False
                                                    },
                                                    'ul-ReferenceSignalsPUSCH': {
                                                        'groupHoppingEnabled': True,
                                                        'groupAssignmentPUSCH': 22,
                                                        'sequenceHoppingEnabled': False,
                                                        'cyclicShift': 5
                                                    }
                                                },
                                                'pucch-ConfigCommon': {
                                                    'deltaPUCCH-Shift': 'ds1',
                                                    'nRB-CQI': 98,
                                                    'nCS-AN': 4,
                                                    'n1PUCCH-AN': 2047
                                                },
                                                'soundingRS-UL-ConfigCommon': (
                                                    'setup',
                                                    {
                                                        'srs-BandwidthConfig': 'bw0',
                                                        'srs-SubframeConfig': 'sc4',
                                                        'ackNackSRS-SimultaneousTransmission': True
                                                    }),
                                                'uplinkPowerControlCommon': {
                                                    'p0-NominalPUSCH': -126,
                                                    'alpha': 'al0',
                                                    'p0-NominalPUCCH': -127,
                                                    'deltaFList-PUCCH': {
                                                        'deltaF-PUCCH-Format1': 'deltaF-2',
                                                        'deltaF-PUCCH-Format1b': 'deltaF1',
                                                        'deltaF-PUCCH-Format2': 'deltaF0',
                                                        'deltaF-PUCCH-Format2a': 'deltaF-2',
                                                        'deltaF-PUCCH-Format2b': 'deltaF0'
                                                    },
                                                    'deltaPreambleMsg3': -1
                                                },
                                                'ul-CyclicPrefixLength': 'len1'
                                            },
                                            'ue-TimersAndConstants': {
                                                't300': 'ms100',
                                                't301': 'ms200',
                                                't310': 'ms50',
                                                'n310': 'n2',
                                                't311': 'ms30000',
                                                'n311': 'n2'
                                            },
                                            'freqInfo': {
                                                'additionalSpectrumEmission': 3
                                            },
                                            'timeAlignmentTimerCommon': 'sf500'
                                        }
                                    ),
                                    (
                                        'sib3',
                                        {
                                            'cellReselectionInfoCommon': {
                                                'q-Hyst': 'dB0',
                                                'speedStateReselectionPars': {
                                                    'mobilityStateParameters': {
                                                        't-Evaluation': 's180',
                                                        't-HystNormal': 's180',
                                                        'n-CellChangeMedium': 1,
                                                        'n-CellChangeHigh': 16
                                                    },
                                                    'q-HystSF': {
                                                        'sf-Medium': 'dB-6',
                                                        'sf-High': 'dB-4'
                                                    }
                                                }
                                            },
                                            'cellReselectionServingFreqInfo': {
                                                'threshServingLow': 7,
                                                'cellReselectionPriority': 3
                                            },
                                            'intraFreqCellReselectionInfo': {
                                                'q-RxLevMin': -33,
                                                's-IntraSearch': 0,
                                                'presenceAntennaPort1': False,
                                                'neighCellConfig': (b'\x80', 2),
                                                't-ReselectionEUTRA': 4
                                            }
                                        }
                                    ),
                                    (
                                        'sib4',
                                        {
                                        }
                                    ),
                                    (
                                        'sib5',
                                        {
                                            'interFreqCarrierFreqList': [
                                                {
                                                    'dl-CarrierFreq': 1,
                                                    'q-RxLevMin': -45,
                                                    't-ReselectionEUTRA': 0,
                                                    'threshX-High': 31,
                                                    'threshX-Low': 29,
                                                    'allowedMeasBandwidth': 'mbw6',
                                                    'presenceAntennaPort1': True,
                                                    'neighCellConfig': (b'\x00', 2),
                                                    'q-OffsetFreq': 'dB0'
                                                }
                                            ]
                                        }
                                    ),
                                    (
                                        'sib6',
                                        {
                                            't-ReselectionUTRA': 3
                                        }
                                    ),
                                    (
                                        'sib7',
                                        {
                                            't-ReselectionGERAN': 3
                                        }
                                    ),
                                    (
                                        'sib8',
                                        {
                                            'parameters1XRTT': {
                                                'longCodeState1XRTT': (b'\x01#Eg\x89\x00', 42)
                                            }
                                        }
                                    ),
                                    (
                                        'sib9',
                                        {
                                            'hnb-Name': b'4'
                                        }
                                    ),
                                    (
                                        'sib10',
                                        {
                                            'messageIdentifier': (b'#4', 16),
                                            'serialNumber': (b'\x124', 16),
                                            'warningType': b'2\x12'
                                        }
                                    ),
                                    (
                                        'sib11',
                                        {
                                            'messageIdentifier': (b'g\x88', 16),
                                            'serialNumber': (b'T5', 16),
                                            'warningMessageSegmentType': 'notLastSegment',
                                            'warningMessageSegmentNumber': 19,
                                            'warningMessageSegment': b'\x12'
                                        }
                                    )
                                ]
                            }
                        )
                    }
                )
            )
        }

        encoded = (
            b'<BCCH-DL-SCH-Message><message><c1><systemInformation><critic'
            b'alExtensions><systemInformation-r8><sib-TypeAndInfo><sib2><a'
            b'c-BarringInfo><ac-BarringForEmergency><true /></ac-BarringFo'
            b'rEmergency><ac-BarringForMO-Data><ac-BarringFactor><p95 /></'
            b'ac-BarringFactor><ac-BarringTime><s128 /></ac-BarringTime><a'
            b'c-BarringForSpecialAC>11110</ac-BarringForSpecialAC></ac-Bar'
            b'ringForMO-Data></ac-BarringInfo><radioResourceConfigCommon><'
            b'rach-ConfigCommon><preambleInfo><numberOfRA-Preambles><n24 /'
            b'></numberOfRA-Preambles><preamblesGroupAConfig><sizeOfRA-Pre'
            b'amblesGroupA><n28 /></sizeOfRA-PreamblesGroupA><messageSizeG'
            b'roupA><b144 /></messageSizeGroupA><messagePowerOffsetGroupB>'
            b'<minusinfinity /></messagePowerOffsetGroupB></preamblesGroup'
            b'AConfig></preambleInfo><powerRampingParameters><powerRamping'
            b'Step><dB0 /></powerRampingStep><preambleInitialReceivedTarge'
            b'tPower><dBm-102 /></preambleInitialReceivedTargetPower></pow'
            b'erRampingParameters><ra-SupervisionInfo><preambleTransMax><n'
            b'8 /></preambleTransMax><ra-ResponseWindowSize><sf6 /></ra-Re'
            b'sponseWindowSize><mac-ContentionResolutionTimer><sf48 /></ma'
            b'c-ContentionResolutionTimer></ra-SupervisionInfo><maxHARQ-Ms'
            b'g3Tx>8</maxHARQ-Msg3Tx></rach-ConfigCommon><bcch-Config><mod'
            b'ificationPeriodCoeff><n2 /></modificationPeriodCoeff></bcch-'
            b'Config><pcch-Config><defaultPagingCycle><rf256 /></defaultPa'
            b'gingCycle><nB><twoT /></nB></pcch-Config><prach-Config><root'
            b'SequenceIndex>836</rootSequenceIndex><prach-ConfigInfo><prac'
            b'h-ConfigIndex>33</prach-ConfigIndex><highSpeedFlag><false />'
            b'</highSpeedFlag><zeroCorrelationZoneConfig>10</zeroCorrelati'
            b'onZoneConfig><prach-FreqOffset>64</prach-FreqOffset></prach-'
            b'ConfigInfo></prach-Config><pdsch-ConfigCommon><referenceSign'
            b'alPower>-60</referenceSignalPower><p-b>2</p-b></pdsch-Config'
            b'Common><pusch-ConfigCommon><pusch-ConfigBasic><n-SB>1</n-SB>'
            b'<hoppingMode><interSubFrame /></hoppingMode><pusch-HoppingOf'
            b'fset>10</pusch-HoppingOffset><enable64QAM><false /></enable6'
            b'4QAM></pusch-ConfigBasic><ul-ReferenceSignalsPUSCH><groupHop'
            b'pingEnabled><true /></groupHoppingEnabled><groupAssignmentPU'
            b'SCH>22</groupAssignmentPUSCH><sequenceHoppingEnabled><false '
            b'/></sequenceHoppingEnabled><cyclicShift>5</cyclicShift></ul-'
            b'ReferenceSignalsPUSCH></pusch-ConfigCommon><pucch-ConfigComm'
            b'on><deltaPUCCH-Shift><ds1 /></deltaPUCCH-Shift><nRB-CQI>98</'
            b'nRB-CQI><nCS-AN>4</nCS-AN><n1PUCCH-AN>2047</n1PUCCH-AN></puc'
            b'ch-ConfigCommon><soundingRS-UL-ConfigCommon><setup><srs-Band'
            b'widthConfig><bw0 /></srs-BandwidthConfig><srs-SubframeConfig'
            b'><sc4 /></srs-SubframeConfig><ackNackSRS-SimultaneousTransmi'
            b'ssion><true /></ackNackSRS-SimultaneousTransmission></setup>'
            b'</soundingRS-UL-ConfigCommon><uplinkPowerControlCommon><p0-N'
            b'ominalPUSCH>-126</p0-NominalPUSCH><alpha><al0 /></alpha><p0-'
            b'NominalPUCCH>-127</p0-NominalPUCCH><deltaFList-PUCCH><deltaF'
            b'-PUCCH-Format1><deltaF-2 /></deltaF-PUCCH-Format1><deltaF-PU'
            b'CCH-Format1b><deltaF1 /></deltaF-PUCCH-Format1b><deltaF-PUCC'
            b'H-Format2><deltaF0 /></deltaF-PUCCH-Format2><deltaF-PUCCH-Fo'
            b'rmat2a><deltaF-2 /></deltaF-PUCCH-Format2a><deltaF-PUCCH-For'
            b'mat2b><deltaF0 /></deltaF-PUCCH-Format2b></deltaFList-PUCCH>'
            b'<deltaPreambleMsg3>-1</deltaPreambleMsg3></uplinkPowerContro'
            b'lCommon><ul-CyclicPrefixLength><len1 /></ul-CyclicPrefixLeng'
            b'th></radioResourceConfigCommon><ue-TimersAndConstants><t300>'
            b'<ms100 /></t300><t301><ms200 /></t301><t310><ms50 /></t310><'
            b'n310><n2 /></n310><t311><ms30000 /></t311><n311><n2 /></n311'
            b'></ue-TimersAndConstants><freqInfo><additionalSpectrumEmissi'
            b'on>3</additionalSpectrumEmission></freqInfo><timeAlignmentTi'
            b'merCommon><sf500 /></timeAlignmentTimerCommon></sib2><sib3><'
            b'cellReselectionInfoCommon><q-Hyst><dB0 /></q-Hyst><speedStat'
            b'eReselectionPars><mobilityStateParameters><t-Evaluation><s18'
            b'0 /></t-Evaluation><t-HystNormal><s180 /></t-HystNormal><n-C'
            b'ellChangeMedium>1</n-CellChangeMedium><n-CellChangeHigh>16</'
            b'n-CellChangeHigh></mobilityStateParameters><q-HystSF><sf-Med'
            b'ium><dB-6 /></sf-Medium><sf-High><dB-4 /></sf-High></q-HystS'
            b'F></speedStateReselectionPars></cellReselectionInfoCommon><c'
            b'ellReselectionServingFreqInfo><threshServingLow>7</threshSer'
            b'vingLow><cellReselectionPriority>3</cellReselectionPriority>'
            b'</cellReselectionServingFreqInfo><intraFreqCellReselectionIn'
            b'fo><q-RxLevMin>-33</q-RxLevMin><s-IntraSearch>0</s-IntraSear'
            b'ch><presenceAntennaPort1><false /></presenceAntennaPort1><ne'
            b'ighCellConfig>10</neighCellConfig><t-ReselectionEUTRA>4</t-R'
            b'eselectionEUTRA></intraFreqCellReselectionInfo></sib3><sib4 '
            b'/><sib5><interFreqCarrierFreqList><InterFreqCarrierFreqInfo>'
            b'<dl-CarrierFreq>1</dl-CarrierFreq><q-RxLevMin>-45</q-RxLevMi'
            b'n><t-ReselectionEUTRA>0</t-ReselectionEUTRA><threshX-High>31'
            b'</threshX-High><threshX-Low>29</threshX-Low><allowedMeasBand'
            b'width><mbw6 /></allowedMeasBandwidth><presenceAntennaPort1><'
            b'true /></presenceAntennaPort1><neighCellConfig>00</neighCell'
            b'Config><q-OffsetFreq><dB0 /></q-OffsetFreq></InterFreqCarrie'
            b'rFreqInfo></interFreqCarrierFreqList></sib5><sib6><t-Reselec'
            b'tionUTRA>3</t-ReselectionUTRA></sib6><sib7><t-ReselectionGER'
            b'AN>3</t-ReselectionGERAN></sib7><sib8><parameters1XRTT><long'
            b'CodeState1XRTT>000000010010001101000101011001111000100100</l'
            b'ongCodeState1XRTT></parameters1XRTT></sib8><sib9><hnb-Name>3'
            b'4</hnb-Name></sib9><sib10><messageIdentifier>001000110011010'
            b'0</messageIdentifier><serialNumber>0001001000110100</serialN'
            b'umber><warningType>3212</warningType></sib10><sib11><message'
            b'Identifier>0110011110001000</messageIdentifier><serialNumber'
            b'>0101010000110101</serialNumber><warningMessageSegmentType><'
            b'notLastSegment /></warningMessageSegmentType><warningMessage'
            b'SegmentNumber>19</warningMessageSegmentNumber><warningMessag'
            b'eSegment>12</warningMessageSegment></sib11></sib-TypeAndInfo'
            b'></systemInformation-r8></criticalExtensions></systemInforma'
            b'tion></c1></message></BCCH-DL-SCH-Message>'
        )

        self.assert_encode_decode_string(rrc, 'BCCH-DL-SCH-Message', decoded, encoded)

    def test_all_types(self):
        foo = asn1tools.compile_files('tests/files/all_types.asn', 'xer')

        datas = [
            ('Boolean',        True, b'<Boolean><true /></Boolean>'),
            ('Integer',         127, b'<Integer>127</Integer>'),
            ('Integer',           0, b'<Integer>0</Integer>'),
            ('Integer',        -128, b'<Integer>-128</Integer>'),
            ('Sequence',         {}, b'<Sequence />'),
            ('Set',              {}, b'<Set />'),
            ('Sequence2',  {'a': 1}, b'<Sequence2><a>1</a></Sequence2>'),
            ('Set2',       {'a': 2}, b'<Set2><a>2</a></Set2>'),
            ('Ia5string',     'bar', b'<Ia5string>bar</Ia5string>'),
            ('SetOf',            [], b'<SetOf />')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_repr_all_types(self):
        foo = asn1tools.compile_files('tests/files/all_types.asn',
                                      'xer')

        self.assertEqual(repr(foo.types['Boolean']), 'Boolean(Boolean)')
        self.assertEqual(repr(foo.types['Integer']), 'Integer(Integer)')
        self.assertEqual(repr(foo.types['Real']), 'Real(Real)')
        self.assertEqual(repr(foo.types['Bitstring']), 'BitString(Bitstring)')
        self.assertEqual(repr(foo.types['Octetstring']),
                         'OctetString(Octetstring)')
        self.assertEqual(repr(foo.types['Null']), 'Null(Null)')
        self.assertEqual(repr(foo.types['Objectidentifier']),
                         'ObjectIdentifier(Objectidentifier)')
        self.assertEqual(repr(foo.types['Enumerated']),
                         'Enumerated(Enumerated)')
        self.assertEqual(repr(foo.types['Utf8string']),
                         'UTF8String(Utf8string)')
        self.assertEqual(repr(foo.types['Sequence']), 'Sequence(Sequence, [])')
        self.assertEqual(repr(foo.types['Set']), 'Set(Set, [])')
        self.assertEqual(repr(foo.types['Sequence2']),
                         'Sequence(Sequence2, [Integer(a)])')
        self.assertEqual(repr(foo.types['Set2']), 'Set(Set2, [Integer(a)])')
        self.assertEqual(repr(foo.types['Numericstring']),
                         'NumericString(Numericstring)')
        self.assertEqual(repr(foo.types['Printablestring']),
                         'PrintableString(Printablestring)')
        self.assertEqual(repr(foo.types['Ia5string']), 'IA5String(Ia5string)')
        self.assertEqual(repr(foo.types['Universalstring']),
                         'UniversalString(Universalstring)')
        self.assertEqual(repr(foo.types['Visiblestring']),
                         'VisibleString(Visiblestring)')
        self.assertEqual(repr(foo.types['Generalstring']),
                         'GeneralString(Generalstring)')
        self.assertEqual(repr(foo.types['Bmpstring']),
                         'BMPString(Bmpstring)')
        self.assertEqual(repr(foo.types['Teletexstring']),
                         'TeletexString(Teletexstring)')
        self.assertEqual(repr(foo.types['Utctime']), 'UTCTime(Utctime)')
        self.assertEqual(repr(foo.types['SequenceOf']),
                         'SequenceOf(SequenceOf, Integer(INTEGER))')
        self.assertEqual(repr(foo.types['SetOf']),
                         'SetOf(SetOf, Integer(INTEGER))')
        self.assertEqual(repr(foo.types['GeneralizedTime1']),
                         'GeneralizedTime(GeneralizedTime1)')
        self.assertEqual(repr(foo.types['Choice']),
                         'Choice(Choice, [Integer(a)])')
        self.assertEqual(repr(foo.types['Any']), 'Any(Any)')
        self.assertEqual(repr(foo.types['Sequence12']),
                         'Sequence(Sequence12, [SequenceOf(a, Recursive(Sequence12))])')

    def test_all_types_automatic_tags(self):
        foo = asn1tools.compile_files(
            'tests/files/all_types_automatic_tags.asn', 'xer')

        datas = [
            ('Sequence3',
             {'a': 1, 'c': 2,'d': True},
             b'<Sequence3><a>1</a><c>2</c><d><true /></d></Sequence3>')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo, type_name, decoded, encoded)

    def test_indent(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= SEQUENCE { "
            "  a SEQUENCE { "
            "    b BOOLEAN, "
            "    c INTEGER "
            "  } "
            "} "
            "END",
            'xer')

        datas = [
            ('A',
             {'a': {'b': True, 'c': 5}},
             (b'<A>\n'
              b'    <a>\n'
              b'        <b>\n'
              b'            <true />\n'
              b'        </b>\n'
              b'        <c>5</c>\n'
              b'    </a>\n'
              b'</A>\n'))
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode_string(foo,
                                             type_name,
                                             decoded,
                                             encoded,
                                             indent=4)

    def test_issue_34(self):
        """Test that a choice type with a recursive member can be compiled and
        used.

        """

        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= BEGIN "
            "A ::= SEQUENCE { "
            "  a B "
            "} "
            "B ::= CHOICE { "
            "  b A, "
            "  c NULL "
            "} "
            "END ",
            'xer')

        decoded = {'a': ('b', {'a': ('c', None)})}
        encoded = b'<A><a><b><a><c /></a></b></a></A>'
        self.assert_encode_decode_string(foo,
                                         'A',
                                         decoded,
                                         encoded)


if __name__ == '__main__':
    unittest.main()



EXPECTED = {'X680': {'extensibility-implied': False,
          'imports': {},
          'object-classes': {'ATTRIBUTE-E-2-15': {'members': [{'name': '&AttributeType',
                                                               'type': 'OpenType'},
                                                              {'name': '&attributeId',
                                                               'type': 'OBJECT '
                                                                       'IDENTIFIER'}]}},
          'object-sets': {},
          'tags': 'AUTOMATIC',
          'types': {'A': {'type': 'NULL'},
                    'A-19-5': {'type': 'ENUMERATED',
                               'values': [('a', 0), ('b', 3), None, ('c', 1)]},
                    'A-19-6': {'type': 'ENUMERATED',
                               'values': [('a', 0), ('b', 1), None, ('c', 2)]},
                    'A-G-1-4-1': {'members': [{'name': 'a', 'type': 'INTEGER'},
                                              None],
                                  'type': 'SEQUENCE'},
                    'A-G-1-4-2': {'members': [{'name': 'a', 'type': 'INTEGER'},
                                              None,
                                              {'name': 'b', 'type': 'BOOLEAN'},
                                              {'name': 'c', 'type': 'INTEGER'}],
                                  'type': 'SEQUENCE'},
                    'A-G-1-4-3': {'members': [{'name': 'a', 'type': 'INTEGER'},
                                              None,
                                              {'name': 'b', 'type': 'BOOLEAN'},
                                              {'name': 'c', 'type': 'INTEGER'},
                                              {'members': [{'name': 'e',
                                                            'type': 'INTEGER'},
                                                           None,
                                                           None,
                                                           {'name': 'f',
                                                            'type': 'IA5String'}],
                                               'name': 'd',
                                               'type': 'SEQUENCE'}],
                                  'type': 'SEQUENCE'},
                    'A-G-1-4-4': {'members': [{'name': 'a', 'type': 'INTEGER'},
                                              None,
                                              {'name': 'b', 'type': 'BOOLEAN'},
                                              {'name': 'c', 'type': 'INTEGER'},
                                              {'members': [{'name': 'e',
                                                            'type': 'INTEGER'},
                                                           None,
                                                           {'name': 'g',
                                                            'optional': True,
                                                            'type': 'BOOLEAN'},
                                                           {'name': 'h',
                                                            'type': 'BMPString'},
                                                           None,
                                                           {'name': 'f',
                                                            'type': 'IA5String'}],
                                               'name': 'd',
                                               'type': 'SEQUENCE'}],
                                  'type': 'SEQUENCE'},
                    'A-G-1-7-1': {'members': [{'name': 'a', 'type': 'INTEGER'},
                                              None],
                                  'type': 'SEQUENCE'},
                    'A-G-1-7-2': {'members': [{'name': 'a', 'type': 'INTEGER'},
                                              None,
                                              [{'name': 'b', 'type': 'BOOLEAN'},
                                               {'name': 'c',
                                                'type': 'INTEGER'}]],
                                  'type': 'SEQUENCE'},
                    'A-G-1-7-3': {'members': [{'name': 'a', 'type': 'INTEGER'},
                                              None,
                                              [{'name': 'b', 'type': 'BOOLEAN'},
                                               {'name': 'c',
                                                'type': 'INTEGER'}],
                                              {'members': [{'name': 'e',
                                                            'type': 'INTEGER'},
                                                           None,
                                                           None,
                                                           {'name': 'f',
                                                            'type': 'IA5String'}],
                                               'name': 'd',
                                               'type': 'SEQUENCE'}],
                                  'type': 'SEQUENCE'},
                    'A-G-1-7-4': {'members': [{'name': 'a', 'type': 'INTEGER'},
                                              None,
                                              [{'name': 'b', 'type': 'BOOLEAN'},
                                               {'name': 'c',
                                                'type': 'INTEGER'}],
                                              {'members': [{'name': 'e',
                                                            'type': 'INTEGER'},
                                                           None,
                                                           [{'name': 'g',
                                                             'optional': True,
                                                             'type': 'BOOLEAN'},
                                                            {'name': 'h',
                                                             'type': 'BMPString'}],
                                                           None,
                                                           {'name': 'f',
                                                            'type': 'IA5String'}],
                                               'name': 'd',
                                               'type': 'SEQUENCE'}],
                                  'type': 'SEQUENCE'},
                    'A-G-4-1-2': {'restricted-to': [('MIN', 'MAX'),
                                                    None,
                                                    (1, 10)],
                                  'type': 'INTEGER'},
                    'A1-G-4-2-3': {'restricted-to': [(1, 32), None, (33, 128)],
                                   'type': 'INTEGER'},
                    'A3-G-4-2-3': {'type': 'INTEGER'},
                    'ABEnvelope-E-4-a': {'type': 'Envelope-E-4',
                                         'with-components': [None,
                                                             ('typeB',
                                                              'PRESENT'),
                                                             ('typeC',
                                                              'ABSENT')]},
                    'ABEnvelope-E-4-b': {'type': 'Envelope-E-4',
                                         'with-components': [['typeA'],
                                                             ['typeB']]},
                    'ACEnvelope-E-4-a': {'type': 'Envelope-E-4',
                                         'with-components': [None,
                                                             ('typeB',
                                                              'ABSENT'),
                                                             ('typeC',
                                                              'PRESENT')]},
                    'ACEnvelope-E-4-b': {'type': 'Envelope-E-4',
                                         'with-components': [['typeA'],
                                                             ['typeC']]},
                    'AcmeBadgeNumber-E-2-12': {'tag': {'class': 'PRIVATE',
                                                       'number': 2},
                                               'type': 'INTEGER'},
                    'Address-E-4': {'size': [(1, 6)], 'type': 'Text-block-E-4'},
                    'AngleInRadians-E-2-4': {'type': 'REAL'},
                    'App-X-Real-E-2-4': {'type': 'REAL',
                                         'with-components': [('mantissa',
                                                              (-16777215,
                                                               16777215)),
                                                             ('base', 2),
                                                             ('exponent',
                                                              (-125, 128))]},
                    'AtomicNumber-E-4': {'restricted-to': [(1, 104)],
                                         'type': 'INTEGER'},
                    'Attribute-E-2-15': {'members': [{'name': 'attributeID',
                                                      'table': None,
                                                      'type': 'ATTRIBUTE-E-2-15.&attributeId'},
                                                     {'name': 'attributeValue',
                                                      'table': None,
                                                      'type': 'ATTRIBUTE-E-2-15.&AttributeType'}],
                                         'type': 'SEQUENCE'},
                    'Audio': {'type': 'NULL'},
                    'B': {'type': 'NULL'},
                    'B-19-6': {'type': 'ENUMERATED',
                               'values': [('a', 1),
                                          ('b', 2),
                                          ('c', 0),
                                          None,
                                          ('d', 3)]},
                    'B1-G-4-2-3': {'restricted-to': [(1, 128)],
                                   'type': 'A1-G-4-2-3'},
                    'B2-G-4-2-3': {'restricted-to': [(1, 16)],
                                   'type': 'A1-G-4-2-3'},
                    'BitField-E-2-5': {'size': [12], 'type': 'BIT STRING'},
                    'C': {'type': 'NULL'},
                    'C-19-6': {'type': 'ENUMERATED',
                               'values': [('a', 0),
                                          ('b', 1),
                                          None,
                                          ('c', 3),
                                          ('d', 4)]},
                    'C0-F-2-6': {'from': [('{', '{')], 'type': 'BMPString'},
                    'C1-F-2-6': {'from': [('{', '{')], 'type': 'BMPString'},
                    'Certificate': {'type': 'NULL'},
                    'CheckingAccountBalance-E-2-2': {'type': 'INTEGER'},
                    'ChildInformation-E-1-1': {'members': [{'name': 'name',
                                                            'type': 'Name-E-1-1'},
                                                           {'name': 'dateOfBirth',
                                                            'type': 'Date-E-1-1'}],
                                               'type': 'SET'},
                    'Credentials-E-2-10': {'members': [{'name': 'userName',
                                                        'type': 'VisibleString'},
                                                       {'name': 'password',
                                                        'type': 'VisibleString'},
                                                       {'name': 'accountNumber',
                                                        'type': 'INTEGER'}],
                                           'type': 'SEQUENCE'},
                    'CustomerAttribute-E-2-12-a': {'members': [{'name': 'name',
                                                                'tag': {'number': 0},
                                                                'type': 'VisibleString'},
                                                               {'name': 'mailingAddress',
                                                                'tag': {'number': 1},
                                                                'type': 'VisibleString'},
                                                               {'name': 'accountNumber',
                                                                'tag': {'number': 2},
                                                                'type': 'INTEGER'},
                                                               {'name': 'balanceDue',
                                                                'tag': {'number': 3},
                                                                'type': 'INTEGER'}],
                                                   'type': 'CHOICE'},
                    'CustomerAttribute-E-2-12-b': {'members': [{'name': 'name',
                                                                'tag': {'kind': 'IMPLICIT',
                                                                        'number': 0},
                                                                'type': 'VisibleString'},
                                                               {'name': 'mailingAddress',
                                                                'tag': {'kind': 'IMPLICIT',
                                                                        'number': 1},
                                                                'type': 'VisibleString'},
                                                               {'name': 'accountNumber',
                                                                'tag': {'kind': 'IMPLICIT',
                                                                        'number': 2},
                                                                'type': 'INTEGER'},
                                                               {'name': 'balanceDue',
                                                                'tag': {'kind': 'IMPLICIT',
                                                                        'number': 3},
                                                                'type': 'INTEGER'}],
                                                   'type': 'CHOICE'},
                    'CustomerRecord-E-2-12-a': {'members': [{'name': 'name',
                                                             'tag': {'number': 0},
                                                             'type': 'VisibleString'},
                                                            {'name': 'mailingAddress',
                                                             'tag': {'number': 1},
                                                             'type': 'VisibleString'},
                                                            {'name': 'accountNumber',
                                                             'tag': {'number': 2},
                                                             'type': 'INTEGER'},
                                                            {'name': 'balanceDue',
                                                             'tag': {'number': 3},
                                                             'type': 'INTEGER'}],
                                                'type': 'SET'},
                    'CustomerRecord-E-2-12-b': {'members': [{'name': 'name',
                                                             'tag': {'kind': 'IMPLICIT',
                                                                     'number': 0},
                                                             'type': 'VisibleString'},
                                                            {'name': 'mailingAddress',
                                                             'tag': {'kind': 'IMPLICIT',
                                                                     'number': 1},
                                                             'type': 'VisibleString'},
                                                            {'name': 'accountNumber',
                                                             'tag': {'kind': 'IMPLICIT',
                                                                     'number': 2},
                                                             'type': 'INTEGER'},
                                                            {'name': 'balanceDue',
                                                             'tag': {'kind': 'IMPLICIT',
                                                                     'number': 3},
                                                             'type': 'INTEGER'}],
                                                'type': 'SET'},
                    'D': {'type': 'NULL'},
                    'D-19-5': {'type': 'ENUMERATED',
                               'values': [('a', 0), ('b', 1), None, ('c', 2)]},
                    'D-19-6': {'type': 'ENUMERATED',
                               'values': [('a', 0), ('z', 25), None, ('d', 1)]},
                    'Date-E-1-1': {'tag': {'class': 'APPLICATION', 'number': 3},
                                   'type': 'VisibleString'},
                    'DayOfTheMonth-E-2-2': {'type': 'INTEGER'},
                    'DayOfTheMonth-E-2-2-a': {'restricted-to': ['first',
                                                                'last'],
                                              'type': 'INTEGER'},
                    'DayOfTheMonth-E-2-2-b': {'restricted-to': [('first',
                                                                 'last')],
                                              'type': 'INTEGER'},
                    'DayOfTheWeek-E-2-3': {'type': 'ENUMERATED',
                                           'values': [('sunday', 0),
                                                      ('monday', 1),
                                                      ('tuesday', 2),
                                                      ('wednesday', 3),
                                                      ('thursday', 4),
                                                      ('friday', 5),
                                                      ('saturday', 6)]},
                    'DaysOfTheWeek-E-2-5-a': {'named-bits': [('sunday', '0'),
                                                             ('monday', '1'),
                                                             ('tuesday', '2'),
                                                             ('wednesday', '3'),
                                                             ('thursday', '4'),
                                                             ('friday', '5'),
                                                             ('saturday', '6')],
                                              'size': [(0, 7)],
                                              'type': 'BIT STRING'},
                    'DaysOfTheWeek-E-2-5-b': {'named-bits': [('sunday', '0'),
                                                             ('monday', '1'),
                                                             ('tuesday', '2'),
                                                             ('wednesday', '3'),
                                                             ('thursday', '4'),
                                                             ('friday', '5'),
                                                             ('saturday', '6')],
                                              'size': [7],
                                              'type': 'BIT STRING'},
                    'E': {'type': 'NULL'},
                    'Employed-E-2-1': {'type': 'BOOLEAN'},
                    'EmployeeNumber-E-1-1': {'tag': {'class': 'APPLICATION',
                                                     'number': 2},
                                             'type': 'INTEGER'},
                    'Envelope-E-4': {'members': [{'name': 'typeA',
                                                  'type': 'TypeA'},
                                                 {'name': 'typeB',
                                                  'optional': True,
                                                  'type': 'TypeB'},
                                                 {'name': 'typeC',
                                                  'optional': True,
                                                  'type': 'TypeC'}],
                                     'type': 'SET'},
                    'FileIdentifier-E-2-13-a': {'members': [{'name': 'relativeName',
                                                             'type': 'VisibleString'},
                                                            {'name': 'absoluteName',
                                                             'type': 'VisibleString'},
                                                            {'name': 'serialNumber',
                                                             'type': 'INTEGER'}],
                                                'type': 'CHOICE'},
                    'FileIdentifier-E-2-13-b': {'members': [{'name': 'relativeName',
                                                             'type': 'VisibleString'},
                                                            {'name': 'absoluteName',
                                                             'type': 'VisibleString'},
                                                            None,
                                                            None],
                                                'type': 'CHOICE'},
                    'FileIdentifier-E-2-13-c': {'members': [{'name': 'relativeName',
                                                             'type': 'VisibleString'},
                                                            {'name': 'absoluteName',
                                                             'type': 'VisibleString'},
                                                            None,
                                                            {'name': 'serialNumber',
                                                             'type': 'INTEGER'},
                                                            None],
                                                'type': 'CHOICE'},
                    'FileIdentifier-E-2-13-d': {'members': [{'name': 'relativeName',
                                                             'type': 'VisibleString'},
                                                            {'name': 'absoluteName',
                                                             'type': 'VisibleString'},
                                                            None,
                                                            {'name': 'serialNumber',
                                                             'type': 'INTEGER'},
                                                            [{'name': 'vendorSpecific',
                                                              'type': 'VendorExt'},
                                                             {'name': 'unidentified',
                                                              'type': 'NULL'}],
                                                            None],
                                                'type': 'CHOICE'},
                    'FileName-E-2-12': {'members': [{'name': 'directoryName',
                                                     'type': 'VisibleString'},
                                                    {'name': 'directoryRelativeFileName',
                                                     'type': 'VisibleString'}],
                                        'tag': {'class': 'APPLICATION',
                                                'number': 8},
                                        'type': 'SEQUENCE'},
                    'First-half-E-4': {'restricted-to': [{'type': 'First-quarter-E-4'},
                                                         {'type': 'Second-quarter-E-4'}],
                                       'type': 'Months-E-4'},
                    'First-quarter-E-4': {'restricted-to': ['january',
                                                            'february',
                                                            'march'],
                                          'type': 'Months-E-4'},
                    'Fourth-quarter-E-4': {'restricted-to': ['october',
                                                             'november',
                                                             'december'],
                                           'type': 'Months-E-4'},
                    'FurtherTestPDU-E-4': {'type': 'TestPDU-E-4',
                                           'with-components': [None,
                                                               ['beta',
                                                                [{'size': [5,
                                                                           12]}],
                                                                'PRESENT']]},
                    'G3FacsimilePage-E-2-5': {'type': 'BIT STRING'},
                    'G4FacsimileImage-E-2-6': {'type': 'OCTET STRING'},
                    'Greeting-E-2-13-a': {'members': [{'name': 'postCard',
                                                       'type': 'VisibleString'},
                                                      None,
                                                      None],
                                          'type': 'CHOICE'},
                    'Greeting-E-2-13-b': {'members': [{'name': 'postCard',
                                                       'type': 'VisibleString'},
                                                      None,
                                                      [{'name': 'audio',
                                                        'type': 'Audio'},
                                                       {'name': 'video',
                                                        'type': 'Video'}],
                                                      None],
                                          'type': 'CHOICE'},
                    'Greeting-E-2-13-c': {'members': [{'name': 'postCard',
                                                       'type': 'VisibleString'},
                                                      {'name': 'recording',
                                                       'type': 'Voice-E-2-13'}],
                                          'tag': {'class': 'APPLICATION',
                                                  'number': 12},
                                          'type': 'CHOICE'},
                    'Keywords-E-2-11': {'element': {'type': 'VisibleString'},
                                        'type': 'SET OF'},
                    'Keywords2-E-2-11': {'element': {'type': 'VisibleString'},
                                         'type': 'SET OF'},
                    'MaritalStatus-E-2-3-a': {'type': 'ENUMERATED',
                                              'values': [('single', 0),
                                                         ('married', 1)]},
                    'MaritalStatus-E-2-3-b': {'type': 'ENUMERATED',
                                              'values': [('single', 0),
                                                         ('married', 1),
                                                         None,
                                                         ('widowed', 2)]},
                    'MaritalStatus-E-2-3-c': {'type': 'ENUMERATED',
                                              'values': [('single', 0),
                                                         ('married', 1),
                                                         None,
                                                         ('widowed', 2),
                                                         ('divorced', 3)]},
                    'Months-E-4': {'type': 'ENUMERATED',
                                   'values': [('january', 1),
                                              ('february', 2),
                                              ('march', 3),
                                              ('april', 4),
                                              ('may', 5),
                                              ('june', 6),
                                              ('july', 7),
                                              ('august', 8),
                                              ('september', 9),
                                              ('october', 10),
                                              ('november', 11),
                                              ('december', 12)]},
                    'Name-E-1-1': {'members': [{'name': 'givenName',
                                                'type': 'VisibleString'},
                                               {'name': 'initial',
                                                'type': 'VisibleString'},
                                               {'name': 'familyName',
                                                'type': 'VisibleString'}],
                                   'tag': {'class': 'APPLICATION', 'number': 1},
                                   'type': 'SEQUENCE'},
                    'NamesOfMemberNations-E-2-10': {'element': {'type': 'VisibleString'},
                                                    'type': 'SEQUENCE OF'},
                    'NamesOfMemberNations2-E-2-10': {'element': {'type': 'VisibleString'},
                                                     'type': 'SEQUENCE OF'},
                    'NamesOfOfficers-E-2-10': {'members': [{'name': 'president',
                                                            'type': 'VisibleString'},
                                                           {'name': 'vicePresident',
                                                            'type': 'VisibleString'},
                                                           {'name': 'secretary',
                                                            'type': 'VisibleString'}],
                                               'type': 'SEQUENCE'},
                    'PDU-E-4': {'members': [{'name': 'alpha',
                                             'type': 'INTEGER'},
                                            {'name': 'beta',
                                             'optional': True,
                                             'type': 'IA5String'},
                                            {'element': {'type': 'Parameter'},
                                             'name': 'gamma',
                                             'type': 'SEQUENCE OF'},
                                            {'name': 'delta',
                                             'type': 'BOOLEAN'}],
                                'type': 'SET'},
                    'PackedBCDString-E-2-8': {'type': 'CHARACTER STRING',
                                              'with-components': [('identification',
                                                                   {'with-components': [('fixed',
                                                                                         'PRESENT')]})]},
                    'Parameter': {'type': 'NULL'},
                    'ParameterList-E-4': {'element': {'type': 'Parameter'},
                                          'size': [(1, 63)],
                                          'type': 'SET OF'},
                    'PatientIdentifier-E-2-9': {'members': [{'name': 'name',
                                                             'type': 'VisibleString'},
                                                            {'members': [{'name': 'room',
                                                                          'type': 'INTEGER'},
                                                                         {'name': 'outPatient',
                                                                          'type': 'NULL'}],
                                                             'name': 'roomNumber',
                                                             'type': 'CHOICE'}],
                                                'type': 'SEQUENCE'},
                    'PersonalStatus-E-2-5': {'named-bits': [('married', '0'),
                                                            ('employed', '1'),
                                                            ('veteran', '2'),
                                                            ('collegeGraduate',
                                                             '3')],
                                             'type': 'BIT STRING'},
                    'PersonnelRecord-E-1-1': {'members': [{'name': 'name',
                                                           'type': 'Name-E-1-1'},
                                                          {'name': 'title',
                                                           'type': 'VisibleString'},
                                                          {'name': 'number',
                                                           'type': 'EmployeeNumber-E-1-1'},
                                                          {'name': 'dateOfHire',
                                                           'type': 'Date-E-1-1'},
                                                          {'name': 'nameOfSpouse',
                                                           'type': 'Name-E-1-1'},
                                                          {'default': [],
                                                           'element': {'type': 'ChildInformation-E-1-1'},
                                                           'name': 'children',
                                                           'type': 'SEQUENCE '
                                                                   'OF'}],
                                              'tag': {'class': 'APPLICATION',
                                                      'number': 0},
                                              'type': 'SET'},
                    'Record-E-2-10-a': {'members': [{'name': 'userName',
                                                     'type': 'VisibleString'},
                                                    {'name': 'password',
                                                     'type': 'VisibleString'},
                                                    {'name': 'accountNumber',
                                                     'type': 'INTEGER'},
                                                    None,
                                                    None],
                                        'type': 'SEQUENCE'},
                    'Record-E-2-10-b': {'members': [{'name': 'userName',
                                                     'type': 'VisibleString'},
                                                    {'name': 'password',
                                                     'type': 'VisibleString'},
                                                    {'name': 'accountNumber',
                                                     'type': 'INTEGER'},
                                                    None,
                                                    [{'name': 'lastLoggedIn',
                                                      'optional': True,
                                                      'type': 'GeneralizedTime'},
                                                     {'name': 'minutesLastLoggedIn',
                                                      'type': 'INTEGER'}],
                                                    None],
                                        'type': 'SEQUENCE'},
                    'Record-E-2-10-c': {'members': [{'name': 'userName',
                                                     'type': 'VisibleString'},
                                                    {'name': 'password',
                                                     'type': 'VisibleString'},
                                                    {'name': 'accountNumber',
                                                     'type': 'INTEGER'},
                                                    None,
                                                    [{'name': 'lastLoggedIn',
                                                      'optional': True,
                                                      'type': 'GeneralizedTime'},
                                                     {'name': 'minutesLastLoggedIn',
                                                      'type': 'INTEGER'}],
                                                    [{'name': 'certificate',
                                                      'type': 'Certificate'},
                                                     {'name': 'thumb',
                                                      'optional': True,
                                                      'type': 'ThumbPrint'}],
                                                    None],
                                        'type': 'SEQUENCE'},
                    'Second-half-E-4': {'restricted-to': [{'type': 'Third-quarter-E-4'},
                                                          {'type': 'Fourth-quarter-E-4'}],
                                        'type': 'Months-E-4'},
                    'Second-quarter-E-4': {'restricted-to': ['april',
                                                             'may',
                                                             'june'],
                                           'type': 'Months-E-4'},
                    'SmallPrime-E-4-a': {'restricted-to': [2,
                                                           3,
                                                           5,
                                                           7,
                                                           11,
                                                           13,
                                                           17,
                                                           19,
                                                           23,
                                                           29],
                                         'type': 'INTEGER'},
                    'SmallPrime-E-4-b': {'restricted-to': [2, 3, None],
                                         'type': 'INTEGER'},
                    'SmallPrime-E-4-c': {'restricted-to': [2,
                                                           3,
                                                           None,
                                                           5,
                                                           7,
                                                           11],
                                         'type': 'INTEGER'},
                    'SmallPrime-E-4-d': {'restricted-to': [2,
                                                           3,
                                                           None,
                                                           5,
                                                           7,
                                                           11,
                                                           13,
                                                           17,
                                                           19],
                                         'type': 'INTEGER'},
                    'Surname-E-2-6': {'type': 'PrintableString'},
                    'T-14-10': {'members': [{'name': 'a', 'type': 'BOOLEAN'},
                                            {'element': {'type': 'INTEGER'},
                                             'name': 'b',
                                             'type': 'SET OF'}],
                                'type': 'SEQUENCE'},
                    'TestPDU-E-4': {'type': 'PDU-E-4',
                                    'with-components': [None,
                                                        ('delta', 'FALSE'),
                                                        ('alpha',
                                                         ('MIN', '<'))]},
                    'Text-block-E-4': {'element': {'type': 'VisibleString'},
                                       'type': 'SEQUENCE OF'},
                    'Third-quarter-E-4': {'restricted-to': ['july',
                                                            'august',
                                                            'september'],
                                          'type': 'Months-E-4'},
                    'ThumbPrint': {'type': 'NULL'},
                    'TouchToneString-E-4': {'from': [('0', '0'),
                                                     ('1', '1'),
                                                     ('2', '2'),
                                                     ('3', '3'),
                                                     ('4', '4'),
                                                     ('5', '5'),
                                                     ('6', '6'),
                                                     ('7', '7'),
                                                     ('8', '8'),
                                                     ('9', '9'),
                                                     ('*', '*'),
                                                     ('#', '#')],
                                            'size': [(1, 63)],
                                            'type': 'IA5String'},
                    'TypeA': {'type': 'INTEGER'},
                    'TypeB': {'type': 'BIT STRING'},
                    'TypeC': {'type': 'REAL'},
                    'UserName-E-2-11-a': {'members': [{'name': 'personalName',
                                                       'tag': {'number': 0},
                                                       'type': 'VisibleString'},
                                                      {'name': 'organizationName',
                                                       'tag': {'number': 1},
                                                       'type': 'VisibleString'},
                                                      {'name': 'countryName',
                                                       'tag': {'number': 2},
                                                       'type': 'VisibleString'}],
                                          'type': 'SET'},
                    'UserName-E-2-11-b': {'members': [{'name': 'personalName',
                                                       'tag': {'number': 0},
                                                       'type': 'VisibleString'},
                                                      {'name': 'organizationName',
                                                       'optional': True,
                                                       'tag': {'number': 1},
                                                       'type': 'VisibleString'},
                                                      {'name': 'countryName',
                                                       'optional': True,
                                                       'tag': {'number': 2},
                                                       'type': 'VisibleString'}],
                                          'type': 'SET'},
                    'UserName-E-2-11-c': {'members': [{'name': 'personalName',
                                                       'type': 'VisibleString'},
                                                      {'name': 'organizationName',
                                                       'optional': True,
                                                       'type': 'VisibleString'},
                                                      {'name': 'countryName',
                                                       'optional': True,
                                                       'type': 'VisibleString'},
                                                      None,
                                                      None],
                                          'type': 'SET'},
                    'UserName-E-2-11-d': {'members': [{'name': 'personalName',
                                                       'type': 'VisibleString'},
                                                      {'name': 'organizationName',
                                                       'optional': True,
                                                       'type': 'VisibleString'},
                                                      {'name': 'countryName',
                                                       'optional': True,
                                                       'type': 'VisibleString'},
                                                      None,
                                                      [{'name': 'internetEmailAddress',
                                                        'type': 'VisibleString'},
                                                       {'name': 'faxNumber',
                                                        'optional': True,
                                                        'type': 'VisibleString'}],
                                                      None],
                                          'type': 'SET'},
                    'UserName-E-2-11-e': {'members': [{'name': 'personalName',
                                                       'type': 'VisibleString'},
                                                      {'name': 'organizationName',
                                                       'optional': True,
                                                       'type': 'VisibleString'},
                                                      {'name': 'countryName',
                                                       'optional': True,
                                                       'type': 'VisibleString'},
                                                      None,
                                                      [{'name': 'internetEmailAddress',
                                                        'type': 'VisibleString'},
                                                       {'name': 'faxNumber',
                                                        'optional': True,
                                                        'type': 'VisibleString'}],
                                                      [{'name': 'phoneNumber',
                                                        'optional': True,
                                                        'type': 'VisibleString'}],
                                                      None],
                                          'type': 'SET'},
                    'V-E-4': {'type': 'Z-E-4',
                              'with-components': [None,
                                                  ('a', 'ABSENT'),
                                                  ('b', 'ABSENT')]},
                    'VendorExt': {'type': 'OCTET STRING'},
                    'Video': {'type': 'REAL'},
                    'Voice-E-2-13': {'members': [{'name': 'english',
                                                  'type': 'OCTET STRING'},
                                                 {'name': 'swahili',
                                                  'type': 'OCTET STRING'}],
                                     'type': 'CHOICE'},
                    'W-E-4': {'type': 'Z-E-4',
                              'with-components': [None, ('a', 'PRESENT')]},
                    'X-E-4': {'type': 'Z-E-4',
                              'with-components': [('a', 'PRESENT')]},
                    'Y-E-4': {'type': 'Z-E-4',
                              'with-components': [('a', 'ABSENT'),
                                                  ['b'],
                                                  ['c']]},
                    'Z-E-4': {'members': [{'name': 'a', 'type': 'A'},
                                          {'name': 'b', 'type': 'B'},
                                          {'name': 'c', 'type': 'C'},
                                          {'name': 'd', 'type': 'D'},
                                          {'name': 'e', 'type': 'E'}],
                              'type': 'CHOICE'}},
          'values': {'acmeCorp-E-2-10': {'type': 'NamesOfOfficers-E-2-10',
                                         'value': '{'},
                     'badgeNumber-E-2-12': {'type': 'AcmeBadgeNumber-E-2-12',
                                            'value': 2345},
                     'balance-E-2-2': {'type': 'CheckingAccountBalance-E-2-2',
                                       'value': 0},
                     'billClinton-E-2-5': {'type': 'PersonalStatus-E-2-5',
                                           'value': ['married',
                                                     'employed',
                                                     'collegeGraduate']},
                     'body1-E-2-5': {'type': 'G3FacsimilePage-E-2-5',
                                     'value': '0b1101'},
                     'body2-E-2-5': {'type': 'G3FacsimilePage-E-2-5',
                                     'value': '0b1101000'},
                     'dayOfTheMonth-E-2-2-c': {'type': 'DayOfTheMonth-E-2-2',
                                               'value': 4},
                     'file-E-2-13': {'type': 'FileIdentifier-E-2-13-a',
                                     'value': 'serialNumber'},
                     'fileId1-E-2-13-a': {'type': 'FileIdentifier-E-2-13-b',
                                          'value': 'relativeName'},
                     'fileId1-E-2-13-b': {'type': 'FileIdentifier-E-2-13-c',
                                          'value': 'relativeName'},
                     'fileId1-E-2-13-c': {'type': 'FileIdentifier-E-2-13-d',
                                          'value': 'relativeName'},
                     'fileId2-E-2-13-a': {'type': 'FileIdentifier-E-2-13-c',
                                          'value': 'serialNumber'},
                     'fileId2-E-2-13-b': {'type': 'FileIdentifier-E-2-13-d',
                                          'value': 'serialNumber'},
                     'fileId3-E-2-13-a': {'type': 'FileIdentifier-E-2-13-d',
                                          'value': 'unidentified'},
                     'firstDay-E-2-3': {'type': 'DayOfTheWeek-E-2-3',
                                        'value': 'sunday'},
                     'firstTwo-E-2-10': {'type': 'NamesOfMemberNations-E-2-10',
                                         'value': '{'},
                     'firstTwo2-E-2-10': {'type': 'NamesOfMemberNations2-E-2-10',
                                          'value': '{'},
                     'girth-E-2-4': {'type': 'App-X-Real-E-2-4', 'value': '{'},
                     'greekCapitalLetterSigma-E-2-7': {'type': 'BMPString',
                                                       'value': '{'},
                     'hillaryClinton-E-2-5': {'type': 'PersonalStatus-E-2-5',
                                              'value': '0b110100'},
                     'image-E-2-5': {'type': 'G3FacsimilePage-E-2-5',
                                     'value': '0b100110100100001110110'},
                     'image-E-2-6': {'type': 'G4FacsimileImage-E-2-6',
                                     'value': '0x3fe2ebad471005'},
                     'infinity-E-2-7': {'type': 'UTF8String', 'value': '{'},
                     'lastPatient-E-2-9': {'type': 'PatientIdentifier-E-2-9',
                                           'value': '{'},
                     'map1-E-2-5': {'type': 'BitField-E-2-5',
                                    'value': '0b100110100100'},
                     'map2-E-2-5': {'type': 'BitField-E-2-5', 'value': '0x9a4'},
                     'myGreeting-E-2-13': {'type': 'Greeting-E-2-13-c',
                                           'value': 'recording'},
                     'pi-E-2-4-a': {'type': 'REAL', 'value': None},
                     'pi-E-2-4-b': {'type': 'REAL',
                                    'value': '3.141592653589793'},
                     'president-E-2-6': {'type': 'Surname-E-2-6',
                                         'value': 'Clinton'},
                     'property-E-2-7': {'type': 'UTF8String', 'value': '{'},
                     'rightwardsArrow-E-2-7': {'type': 'UTF8String',
                                               'value': '{'},
                     'someASN1Keywords-E-2-11': {'type': 'Keywords-E-2-11',
                                                 'value': '{'},
                     'someASN1Keywords2-E-2-11': {'type': 'Keywords2-E-2-11',
                                                  'value': '{'},
                     'sunnyDaysLastWeek1-E-2-5-a': {'type': 'DaysOfTheWeek-E-2-5-a',
                                                    'value': ['sunday',
                                                              'monday',
                                                              'wednesday']},
                     'sunnyDaysLastWeek1-E-2-5-b': {'type': 'DaysOfTheWeek-E-2-5-b',
                                                    'value': ['sunday',
                                                              'monday',
                                                              'wednesday']},
                     'sunnyDaysLastWeek2-E-2-5-a': {'type': 'DaysOfTheWeek-E-2-5-a',
                                                    'value': '0b1101'},
                     'sunnyDaysLastWeek3-E-2-5-a': {'type': 'DaysOfTheWeek-E-2-5-a',
                                                    'value': '0b1101000'},
                     'sunnyDaysLastWeek3-E-2-5-b': {'type': 'DaysOfTheWeek-E-2-5-b',
                                                    'value': '0b1101000'},
                     'today-E-2-2': {'type': 'DayOfTheMonth-E-2-2',
                                     'value': 'first'},
                     'trailer-E-2-5': {'type': 'BIT STRING',
                                       'value': '0x0123456789abcdef'},
                     'unknown-E-2-2': {'type': 'DayOfTheMonth-E-2-2',
                                       'value': 0},
                     'user-E-2-11-a': {'type': 'UserName-E-2-11-a',
                                       'value': '{'},
                     'user-E-2-11-b': {'type': 'UserName-E-2-11-c',
                                       'value': '{'},
                     'user-E-2-11-c': {'type': 'UserName-E-2-11-d',
                                       'value': '{'},
                     'user-E-2-11-d': {'type': 'UserName-E-2-11-e',
                                       'value': '{'}}}}


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import unittest
from copy import deepcopy
from datetime import datetime
from .utils import Asn1ToolsBaseTest

import asn1tools
from asn1tools.codecs import restricted_utc_time_to_datetime as ut2dt
from asn1tools.codecs import restricted_generalized_time_to_datetime as gt2dt
from asn1tools.compat import timezone
from asn1tools.compat import timedelta

sys.path.append('tests/files/ietf')

from rfc5280 import EXPECTED as RFC5280


class Asn1ToolsDerTest(Asn1ToolsBaseTest):

    maxDiff = None

    def test_external(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= EXTERNAL "
            "END",
            'der')

        datas = [
            ('A',
             {'encoding': ('octet-aligned', b'\x12')},
             b'\x28\x03\x81\x01\x12')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_real(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= REAL "
            "END")

        datas = [
            ('A',                 0.0, b'\x09\x00'),
            ('A',               100.0, b'\x09\x03\x80\x02\x19'),
            ('A',              -100.0, b'\x09\x03\xc0\x02\x19'),
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

        # Decode 100.0 in decimal form (1.e2).
        self.assertEqual(foo.decode('A', b'\x09\x05\x03\x31\x2e\x45\x32'),
                         100.0)

    def test_bit_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= BIT STRING "
            "B ::= BIT STRING { "
            "  a (0), "
            "  b (1), "
            "  c (2) "
            "} "
            "C ::= SEQUENCE { "
            "  a B DEFAULT {b, c}, "
            "  b B DEFAULT '011'B, "
            "  c B DEFAULT '60'H "
            "} "
            "D ::= SEQUENCE { "
            "  a A DEFAULT '00'B, "
            "  b B DEFAULT '00'B "
            "} "
            "END",
            'der')

        datas = [
            ('A',              (b'', 0), b'\x03\x01\x00'),
            ('A',          (b'\x40', 4), b'\x03\x02\x04\x40'),
            ('A',          (b'\x80', 1), b'\x03\x02\x07\x80'),
            ('A',      (b'\x00\x00', 9), b'\x03\x03\x07\x00\x00'),
            ('B',          (b'\x80', 1), b'\x03\x02\x07\x80'),
            ('B',          (b'\xe0', 3), b'\x03\x02\x05\xe0'),
            ('B',          (b'\x01', 8), b'\x03\x02\x00\x01'),
            ('C',
             {'a': (b'\x40', 2), 'b': (b'\x40', 2), 'c': (b'\x40', 2)},
             b'\x30\x0c\x80\x02\x06\x40\x81\x02\x06\x40\x82\x02\x06\x40'),
            ('C',
             {'a': (b'\x60', 3), 'b': (b'\x60', 3), 'c': (b'\x60', 3)},
             b'\x30\x00'),
            ('D',
             {'a': (b'\x00', 2), 'b': (b'\x00', 2)},
             b'\x30\x00')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

        # Default value is not encoded, but part of decoded. Also
        # ignore dangling bits.
        datas = [
            ('C',
             {},
             b'\x30\x00',
             {'a': (b'\x60', 3), 'b': (b'\x60', 3), 'c': (b'\x60', 3)}),
            ('C',
             {'a': (b'\x60', 4), 'b': (b'\x60', 5), 'c': (b'\x60', 6)},
             b'\x30\x00',
             {'a': (b'\x60', 3), 'b': (b'\x60', 3), 'c': (b'\x60', 3)}),
            ('A',     (b'\xff', 1), b'\x03\x02\x07\x80', (b'\x80', 1)),
            ('D',
             {'a': (b'\x00', 3), 'b': (b'\x00', 3)},
             b'\x30\x04\x80\x02\x05\x00',
             {'a': (b'\x00', 3), 'b': (b'\x00', 2)})
        ]

        for type_name, decoded_1, encoded, decoded_2 in datas:
            self.assertEqual(foo.encode(type_name, decoded_1), encoded)
            self.assertEqual(foo.decode(type_name, encoded), decoded_2)

    def test_set(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS IMPLICIT TAGS ::= "
            "BEGIN "
            "A ::= SET { "
            "  a [0] INTEGER, "
            "  b [1] INTEGER "
            "} "
            "B ::= SET { "
            "  b [1] INTEGER, "
            "  a [0] INTEGER "
            "} "
            "END",
            'ber')

        datas = [
            ('A',     {'a': 3, 'b': 4}, b'\x31\x06\x80\x01\x03\x81\x01\x04'),
            ('B',     {'a': 3, 'b': 4}, b'\x31\x06\x80\x01\x03\x81\x01\x04')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_utf8_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UTF8String "
            "END",
            'der')

        datas = [
            ('A',                u'bar', b'\x0c\x03\x62\x61\x72'),
            ('A',           u'a\u1010c', b'\x0c\x05\x61\xe1\x80\x90\x63')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_graphic_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= GraphicString "
            "END")

        datas = [
            ('A', 'f', b'\x19\x01\x66')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_universal_string(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UniversalString "
            "END",
            'der')

        datas = [
            ('A',   'bar', b'\x1c\x0c\x00\x00\x00b\x00\x00\x00a\x00\x00\x00r'),
            ('A',
             u'åäö',
             b'\x1c\x0c\x00\x00\x00\xe5\x00\x00\x00\xe4\x00\x00\x00\xf6')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_utc_time(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= UTCTime "
            "END",
            'der')

        datas = [
            ('A',
             datetime(2018, 1, 22, 13, 0),
             b'\x17\x0d\x31\x38\x30\x31\x32\x32\x31\x33\x30\x30\x30\x30\x5a'),
            ('A',
             datetime(2001, 2, 3, 4, 5, 6),
             b'\x17\x0d\x30\x31\x30\x32\x30\x33\x30\x34\x30\x35\x30\x36\x5a')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_generalized_time(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= GeneralizedTime "
            "END",
            'der')

        datas = [
            ('A',
             datetime(2018, 1, 22, 13, 29),
             b'\x18\x0f\x32\x30\x31\x38\x30\x31\x32\x32\x31\x33\x32\x39\x30'
             b'\x30\x5a'),
            ('A',
             datetime(2018, 1, 22, 13, 0),
             b'\x18\x0f\x32\x30\x31\x38\x30\x31\x32\x32\x31\x33\x30\x30\x30'
             b'\x30\x5a'),
            ('A',
             datetime(2000, 12, 31, 23, 59, 59),
             b'\x18\x0f\x32\x30\x30\x30\x31\x32\x33\x31\x32\x33\x35\x39'
             b'\x35\x39\x5a'),
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn', 'der')

        datas = [
            ('Boolean',                 True, b'\x01\x01\xff'),
            ('Boolean',                False, b'\x01\x01\x00'),
            ('Integer',                32768, b'\x02\x03\x00\x80\x00'),
            ('Integer',                32767, b'\x02\x02\x7f\xff'),
            ('Integer',                  256, b'\x02\x02\x01\x00'),
            ('Integer',                  255, b'\x02\x02\x00\xff'),
            ('Integer',                  128, b'\x02\x02\x00\x80'),
            ('Integer',                  127, b'\x02\x01\x7f'),
            ('Integer',                    1, b'\x02\x01\x01'),
            ('Integer',                    0, b'\x02\x01\x00'),
            ('Integer',                   -1, b'\x02\x01\xff'),
            ('Integer',                 -128, b'\x02\x01\x80'),
            ('Integer',                 -129, b'\x02\x02\xff\x7f'),
            ('Integer',                 -256, b'\x02\x02\xff\x00'),
            ('Integer',               -32768, b'\x02\x02\x80\x00'),
            ('Integer',               -32769, b'\x02\x03\xff\x7f\xff'),
            ('Bitstring',       (b'\x80', 1), b'\x03\x02\x07\x80'),
            ('Octetstring',          b'\x00', b'\x04\x01\x00'),
            ('Octetstring',    127 * b'\x55', b'\x04\x7f' + 127 * b'\x55'),
            ('Octetstring',    128 * b'\xaa', b'\x04\x81\x80' + 128 * b'\xaa'),
            ('Null',                    None, b'\x05\x00'),
            ('Objectidentifier',       '1.2', b'\x06\x01\x2a'),
            ('Enumerated',             'one', b'\x0a\x01\x01'),
            ('Utf8string',             'foo', b'\x0c\x03foo'),
            ('Sequence',                  {}, b'\x30\x00'),
            ('Sequence2',           {'a': 0}, b'\x30\x00'),
            ('Sequence2',           {'a': 1}, b'\x30\x03\x02\x01\x01'),
            ('Sequence13', {'a': [1]}, b'\x30\x05\xa0\x03\x02\x01\x01'),
            ('Sequence13', {'b': [1]}, b'\x30\x05\xa1\x03\x02\x01\x01'),
            ('Set',                       {}, b'\x31\x00'),
            ('Set2',                {'a': 1}, b'\x31\x00'),
            ('Set2',                {'a': 2}, b'\x31\x03\x02\x01\x02'),
            ('Numericstring',          '123', b'\x12\x03123'),
            ('Printablestring',        'foo', b'\x13\x03foo'),
            ('Ia5string',              'bar', b'\x16\x03bar'),
            ('Visiblestring',          'bar', b'\x1a\x03bar'),
            ('Generalstring',          'bar', b'\x1b\x03bar'),
            ('Bmpstring',
             'bar',
             b'\x1e\x06\x00\x62\x00\x61\x00\x72'),
            ('Teletexstring',          'fum', b'\x14\x03fum'),
            ('SequenceOf',                [], b'0\x00'),
            ('SetOf',                     [], b'1\x00')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(all_types, type_name, decoded, encoded)

        all_types.encode('Sequence12', {'a': [{'a': []}]})

        # ToDo: Should return {'a': [{'a': []}]}
        self.assertEqual(
            all_types.decode('Sequence12', b'\x30\x04\xa0\x02\x30\x00'),
            {})

    def test_repr_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'der')

        self.assertEqual(repr(all_types.types['Boolean']), 'Boolean(Boolean)')
        self.assertEqual(repr(all_types.types['Integer']), 'Integer(Integer)')
        self.assertEqual(repr(all_types.types['Real']), 'Real(Real)')
        self.assertEqual(repr(all_types.types['Bitstring']), 'BitString(Bitstring)')
        self.assertEqual(repr(all_types.types['Octetstring']), 'OctetString(Octetstring)')
        self.assertEqual(repr(all_types.types['Null']), 'Null(Null)')
        self.assertEqual(repr(all_types.types['Objectidentifier']),
                         'ObjectIdentifier(Objectidentifier)')
        self.assertEqual(repr(all_types.types['Enumerated']), 'Enumerated(Enumerated)')
        self.assertEqual(repr(all_types.types['Utf8string']), 'UTF8String(Utf8string)')
        self.assertEqual(repr(all_types.types['Sequence']), 'Sequence(Sequence, [])')
        self.assertEqual(repr(all_types.types['Set']), 'Set(Set, [])')
        self.assertEqual(repr(all_types.types['Sequence2']),
                         'Sequence(Sequence2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Set2']), 'Set(Set2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Numericstring']),
                         'NumericString(Numericstring)')
        self.assertEqual(repr(all_types.types['Printablestring']),
                         'PrintableString(Printablestring)')
        self.assertEqual(repr(all_types.types['Ia5string']), 'IA5String(Ia5string)')
        self.assertEqual(repr(all_types.types['Universalstring']),
                         'UniversalString(Universalstring)')
        self.assertEqual(repr(all_types.types['Visiblestring']),
                         'VisibleString(Visiblestring)')
        self.assertEqual(repr(all_types.types['Generalstring']),
                         'GeneralString(Generalstring)')
        self.assertEqual(repr(all_types.types['Bmpstring']),
                         'BMPString(Bmpstring)')
        self.assertEqual(repr(all_types.types['Teletexstring']),
                         'TeletexString(Teletexstring)')
        self.assertEqual(repr(all_types.types['Graphicstring']),
                         'GraphicString(Graphicstring)')
        self.assertEqual(repr(all_types.types['Utctime']), 'UTCTime(Utctime)')
        self.assertEqual(repr(all_types.types['GeneralizedTime1']),
                         'GeneralizedTime(GeneralizedTime1)')
        self.assertEqual(repr(all_types.types['SequenceOf']),
                         'SequenceOf(SequenceOf, Integer())')
        self.assertEqual(repr(all_types.types['SetOf']), 'SetOf(SetOf, Integer())')
        self.assertEqual(repr(all_types.types['Any']), 'Any(Any)')
        self.assertEqual(repr(all_types.types['Sequence12']),
                         'Sequence(Sequence12, [SequenceOf(a, Recursive(Sequence12))])')

    def test_all_types_automatic_tags(self):
        all_types = asn1tools.compile_files(
            'tests/files/all_types_automatic_tags.asn', 'der')

        datas = [
            ('Sequence3',
             {'a': 1, 'c': 2,'d': True},
             b'\x30\x09\x80\x01\x01\x82\x01\x02\x83\x01\xff')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(all_types, type_name, decoded, encoded)

    def test_decode_length(self):
        foo = asn1tools.compile_files('tests/files/foo.asn', 'der')

        # The length can be decoded.
        datas = [
            (b'\x30\x0e\x02\x01\x01\x16\x09Is 1+1=3?',   16),
            (b'\x30\x10\x02\x02\x01\x16\x09Is 1+10=14?', 18),
            (b'\x30\x0d',                                15),
            (b'\x30\x84\x00\x00\x00\xb8',                190),
            (b'\x9f\x1f\x00', 3)
        ]

        for encoded_message, decoded_length in datas:
            length = foo.decode_length(encoded_message)
            self.assertEqual(length, decoded_length)

        # The length cannot be decoded.
        datas = [
            b'0',
            b'',
            b'0\x84\x00\x00\x00'
        ]

        for encoded in datas:
            self.assertIsNone(foo.decode_length(encoded))

    def test_long_tag(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS IMPLICIT TAGS ::= BEGIN "
            "A ::= [31] INTEGER "
            "B ::= [500] INTEGER "
            "END")

        datas = [
            ('A', 1, b'\x9f\x1f\x01\x01'),
            ('B', 1, b'\x9f\x83\x74\x01\x01')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_rfc5280(self):
        rfc5280 = asn1tools.compile_dict(deepcopy(RFC5280), 'der')

        decoded = {
            'tbsCertificate': {
                'version': 'v1',
                'serialNumber': 3578,
                'signature': {
                    'algorithm': '1.2.840.113549.1.1.5',
                    'parameters': b'\x05\x00'
                },
                'issuer': (
                    'rdnSequence',
                    [
                        [{'type': '2.5.4.6',
                          'value': b'\x13\x02\x4a\x50'}],
                        [{'type': '2.5.4.8',
                          'value': b'\x13\x05\x54\x6f\x6b\x79\x6f'}],
                        [{'type': '2.5.4.7',
                          'value': b'\x13\x07\x43\x68\x75\x6f\x2d\x6b\x75'}],
                        [{'type': '2.5.4.10',
                          'value': b'\x13\x08\x46\x72\x61\x6e\x6b\x34\x44\x44'}],
                        [{'type': '2.5.4.11',
                          'value': (b'\x13\x0f\x57\x65\x62\x43\x65\x72\x74\x20\x53'
                                    b'\x75\x70\x70\x6f\x72\x74')}],
                        [{'type': '2.5.4.3',
                          'value': (b'\x13\x0f\x46\x72\x61\x6e\x6b\x34\x44\x44\x20'
                                    b'\x57\x65\x62\x20\x43\x41')}],
                        [{'type': '1.2.840.113549.1.9.1',
                          'value': (b'\x16\x14\x73\x75\x70\x70\x6f\x72\x74\x40\x66'
                                    b'\x72\x61\x6e\x6b\x34\x64\x64\x2e\x63\x6f\x6d')}]
                    ]
                ),
                'validity': {
                    'notAfter': ('utcTime', ut2dt('170821052654Z')),
                    'notBefore': ('utcTime', ut2dt('120822052654Z'))
                },
                'subject': (
                    'rdnSequence',
                    [
                        [{'type': '2.5.4.6',
                          'value': b'\x13\x02\x4a\x50'}],
                        [{'type': '2.5.4.8',
                          'value': b'\x0c\x05\x54\x6f\x6b\x79\x6f'}],
                        [{'type': '2.5.4.10',
                          'value': b'\x0c\x08\x46\x72\x61\x6e\x6b\x34\x44\x44'}],
                        [{'type': '2.5.4.3',
                          'value': (b'\x0c\x0f\x77\x77\x77\x2e\x65\x78\x61\x6d\x70'
                                    b'\x6c\x65\x2e\x63\x6f\x6d')}]
                    ]
                ),
                'subjectPublicKeyInfo': {
                    'algorithm': {
                        'algorithm': '1.2.840.113549.1.1.1',
                        'parameters': b'\x05\x00'
                    },
                    'subjectPublicKey': (b'0H\x02A\x00\x9b\xfcf\x90y\x84B\xbb'
                                         b'\xab\x13\xfd+{\xf8\xde\x15\x12\xe5'
                                         b'\xf1\x93\xe3\x06\x8a{\xb8\xb1\xe1'
                                         b'\x9e&\xbb\x95\x01\xbf\xe70\xedd\x85'
                                         b'\x02\xdd\x15i\xa84\xb0\x06\xec?5<'
                                         b'\x1e\x1b+\x8f\xfa\x8f\x00\x1b\xdf'
                                         b'\x07\xc6\xacS\x07\x02\x03\x01\x00'
                                         b'\x01',
                                         592)
                }
            },
            'signatureAlgorithm': {
                'algorithm': '1.2.840.113549.1.1.5',
                'parameters': b'\x05\x00'
            },
            'signature': (b'\x14\xb6L\xbb\x81y3\xe6q\xa4\xdaQo\xcb\x08\x1d'
                          b'\x8d`\xec\xbc\x18\xc7sGY\xb1\xf2 H\xbba\xfa'
                          b'\xfcM\xad\x89\x8d\xd1!\xeb\xd5\xd8\xe5\xba'
                          b'\xd6\xa66\xfdtP\x83\xb6\x0f\xc7\x1d\xdf}\xe5.\x81'
                          b'\x7fE\xe0\x9f\xe2>y\xee\xd701\xc7 r\xd9X.*\xfe\x12'
                          b'Z4E\xa1\x19\x08|\x89G_J\x95\xbe#!JSr\xda*\x05/.\xc9'
                          b'p\xf6[\xfa\xfd\xdf\xb41\xb2\xc1J\x9c\x06%C\xa1'
                          b'\xe6\xb4\x1e\x7f\x86\x9b\x16@',
                          1024)
        }

        encoded = (
            b'\x30\x82\x02\x12\x30\x82\x01\x7b\x02\x02\x0d\xfa\x30\x0d\x06\x09'
            b'\x2a\x86\x48\x86\xf7\x0d\x01\x01\x05\x05\x00\x30\x81\x9b\x31\x0b'
            b'\x30\x09\x06\x03\x55\x04\x06\x13\x02\x4a\x50\x31\x0e\x30\x0c\x06'
            b'\x03\x55\x04\x08\x13\x05\x54\x6f\x6b\x79\x6f\x31\x10\x30\x0e\x06'
            b'\x03\x55\x04\x07\x13\x07\x43\x68\x75\x6f\x2d\x6b\x75\x31\x11\x30'
            b'\x0f\x06\x03\x55\x04\x0a\x13\x08\x46\x72\x61\x6e\x6b\x34\x44\x44'
            b'\x31\x18\x30\x16\x06\x03\x55\x04\x0b\x13\x0f\x57\x65\x62\x43\x65'
            b'\x72\x74\x20\x53\x75\x70\x70\x6f\x72\x74\x31\x18\x30\x16\x06\x03'
            b'\x55\x04\x03\x13\x0f\x46\x72\x61\x6e\x6b\x34\x44\x44\x20\x57\x65'
            b'\x62\x20\x43\x41\x31\x23\x30\x21\x06\x09\x2a\x86\x48\x86\xf7\x0d'
            b'\x01\x09\x01\x16\x14\x73\x75\x70\x70\x6f\x72\x74\x40\x66\x72\x61'
            b'\x6e\x6b\x34\x64\x64\x2e\x63\x6f\x6d\x30\x1e\x17\x0d\x31\x32\x30'
            b'\x38\x32\x32\x30\x35\x32\x36\x35\x34\x5a\x17\x0d\x31\x37\x30\x38'
            b'\x32\x31\x30\x35\x32\x36\x35\x34\x5a\x30\x4a\x31\x0b\x30\x09\x06'
            b'\x03\x55\x04\x06\x13\x02\x4a\x50\x31\x0e\x30\x0c\x06\x03\x55\x04'
            b'\x08\x0c\x05\x54\x6f\x6b\x79\x6f\x31\x11\x30\x0f\x06\x03\x55\x04'
            b'\x0a\x0c\x08\x46\x72\x61\x6e\x6b\x34\x44\x44\x31\x18\x30\x16\x06'
            b'\x03\x55\x04\x03\x0c\x0f\x77\x77\x77\x2e\x65\x78\x61\x6d\x70\x6c'
            b'\x65\x2e\x63\x6f\x6d\x30\x5c\x30\x0d\x06\x09\x2a\x86\x48\x86\xf7'
            b'\x0d\x01\x01\x01\x05\x00\x03\x4b\x00\x30\x48\x02\x41\x00\x9b\xfc'
            b'\x66\x90\x79\x84\x42\xbb\xab\x13\xfd\x2b\x7b\xf8\xde\x15\x12\xe5'
            b'\xf1\x93\xe3\x06\x8a\x7b\xb8\xb1\xe1\x9e\x26\xbb\x95\x01\xbf\xe7'
            b'\x30\xed\x64\x85\x02\xdd\x15\x69\xa8\x34\xb0\x06\xec\x3f\x35\x3c'
            b'\x1e\x1b\x2b\x8f\xfa\x8f\x00\x1b\xdf\x07\xc6\xac\x53\x07\x02\x03'
            b'\x01\x00\x01\x30\x0d\x06\x09\x2a\x86\x48\x86\xf7\x0d\x01\x01\x05'
            b'\x05\x00\x03\x81\x81\x00\x14\xb6\x4c\xbb\x81\x79\x33\xe6\x71\xa4'
            b'\xda\x51\x6f\xcb\x08\x1d\x8d\x60\xec\xbc\x18\xc7\x73\x47\x59\xb1'
            b'\xf2\x20\x48\xbb\x61\xfa\xfc\x4d\xad\x89\x8d\xd1\x21\xeb\xd5\xd8'
            b'\xe5\xba\xd6\xa6\x36\xfd\x74\x50\x83\xb6\x0f\xc7\x1d\xdf\x7d\xe5'
            b'\x2e\x81\x7f\x45\xe0\x9f\xe2\x3e\x79\xee\xd7\x30\x31\xc7\x20\x72'
            b'\xd9\x58\x2e\x2a\xfe\x12\x5a\x34\x45\xa1\x19\x08\x7c\x89\x47\x5f'
            b'\x4a\x95\xbe\x23\x21\x4a\x53\x72\xda\x2a\x05\x2f\x2e\xc9\x70\xf6'
            b'\x5b\xfa\xfd\xdf\xb4\x31\xb2\xc1\x4a\x9c\x06\x25\x43\xa1\xe6\xb4'
            b'\x1e\x7f\x86\x9b\x16\x40'
        )

        self.assert_encode_decode(rfc5280, 'Certificate', decoded, encoded)


if __name__ == '__main__':
    unittest.main()



"""Permitted alphabet.

"""

import string


NUMERIC_STRING = ' 0123456789'

PRINTABLE_STRING = (string.ascii_uppercase
                    + string.ascii_lowercase
                    + string.digits
                    + " '()+,-./:=?")

IA5_STRING = ''.join([chr(v) for v in range(128)])

VISIBLE_STRING = ''.join([chr(v) for v in range(32, 127)])
